/*** DX11 CBuffer (Constant Buffer) Manager class, Cristiaan Kop	 upd 28-12-2016 ***/
/*** ============================================================================== ***/
/*** NAMESPACE: Crealysm::RENDERER (DX11)											***/
/*** header file for the DX11CBufferMgr class interface					   		    ***/

/** CONSTANT BUFFERS **/
// 16 byte alignment, padding (float4 = 16 bytes, float4x4 = 64 etc.
// separate CBuffers based on update frequency (per render, per object, etc.)
// CBuffers are in B registers "cbuffer cbPerObject : register (b0) (max 14 per pipeline stage = b13)
//   register is also SLOT NUMBER, for CPU linking
// register numbers can be re-used, as long as you don't let a shader use multiple CB's with same register

/** TEXTURES **/
// have registers tx, max. 128 per pipeline stage -> t0 to t127
// registers can be shared when shader doesn't need textures with same register
// usage: PSSetShaderResource -> using a SRV or UAV, no buffer needed like CBuffers

/** SAMPLERSTATES **/
// have registers sx, max. 16m -> s0 to s15
// PSSetSamplerState

#ifndef DX11CBUFFER_MGR_H
#define DX11CBUFFER_MGR_H

#include <d3d11.h>
#include <vector>
#include <atlbase.h>
#include <ceasylogging++.h>

#include "COMMON/crealysm_text.h"

#define MAX_CB 99
#define LOG_UPDATE 60

namespace Crealysm { namespace RENDERER
{

/**************************************************************************************/
/***								CBUFFER STRUCT									***/
/**************************************************************************************/

struct CBUFFER
{
	CBUFFER() { DataSize = 0; }
	
	CComPtr<ID3D11Buffer>	Buffer;
	size_t					DataSize;
};

/**************************************************************************************/
/***							DX11 CBUFFER MGR CLASS								***/
/*** Manages creation and usage (lifetime) of Constant Buffers, DX11				***/
/**************************************************************************************/

class DX11CBufferMgr 
{
public:
	DX11CBufferMgr(ID3D11Device *pDevice);
	~DX11CBufferMgr();

	DX11CBufferMgr(const DX11CBufferMgr& other) = delete;				// copy constructor: not allowed
	DX11CBufferMgr(DX11CBufferMgr&& other) = delete;					// move constructor: not allowed
	DX11CBufferMgr& operator=(const DX11CBufferMgr& other) = delete;	// copy assignment: not allowed
	DX11CBufferMgr& operator=(DX11CBufferMgr&& other) = delete;			// move assignment: not allowed

	ID3D11Buffer* GetCBufferPtr(const uint pHandle)		const;
	
	template<typename T>bool Add(const T *pBufferData, uint &pHandleCreated)
	{
		uint newId = 0;
	
		// get free spot
		if(mFreeList.size() > 0)
		{
			newId = mFreeList.back();
			mFreeList.pop_back();

			mCBuffers.emplace_back();
		}
		else return false;

		// Create the constant buffer
		D3D11_BUFFER_DESC cBufferDesc;
		cBufferDesc.Usage = D3D11_USAGE_DYNAMIC;
		cBufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
		cBufferDesc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
		cBufferDesc.MiscFlags = 0;
		cBufferDesc.StructureByteStride = 0;
		cBufferDesc.ByteWidth = sizeof(T);

		// pointer to initial data
		D3D11_SUBRESOURCE_DATA initData;
		initData.SysMemPitch = 0;
		initData.SysMemSlicePitch = 0;
		initData.pSysMem = pBufferData;

		if(FAILED(mDevice->CreateBuffer(&cBufferDesc, &initData, &mCBuffers[newId].Buffer))) 
		{
			CLOG(ERROR, "DX11RENDERER") << Crealysm::COMMON::error_create_cb.c_str();
			mFreeList.push_back(newId);
			return false;
		}

		pHandleCreated = newId;
		mCBuffers[newId].DataSize = sizeof(T);

		CLOG(DEBUG, "DX11RENDERER") << Crealysm::COMMON::debug_cbuffer_added.c_str() << newId;
		return true;
	}

	template<typename T>bool Update(ID3D11DeviceContext *pDeviceContext, const uint pHandle, const T *pBufferData)
	{
		static uint logCount = 0;
		
		if(pHandle >= mCBuffers.size()) return false;
		if(sizeof(T) != mCBuffers[pHandle].DataSize) return false;

		D3D11_MAPPED_SUBRESOURCE mappedRes;
		if(FAILED(pDeviceContext->Map(mCBuffers[pHandle].Buffer, 0, D3D11_MAP_WRITE_DISCARD, 0, &mappedRes)))
		{
			CLOG(ERROR, "DX11RENDERER") << Crealysm::COMMON::error_update_cb.c_str();
			return false;
		}
		memcpy(mappedRes.pData, pBufferData, sizeof(T));
		pDeviceContext->Unmap(mCBuffers[pHandle].Buffer, 0);

/*		++logCount;
		if(logCount == LOG_UPDATE)
		{
			CLOG(DEBUG, "DX11RENDERER") << debug_cbuffer_updated.c_str() << pHandle << " (logging each " << LOG_UPDATE << " updates)";
			logCount = 0;
		}*/
		return true;
	}

	bool Release(const uint pHandle);
		
private:
	std::vector<uint>		mFreeList;
	std::vector<CBUFFER>	mCBuffers;
	
	CComPtr<ID3D11Device>	mDevice;
};

}
}

#endif