/*** DX11 Shader Manager class, Cristiaan Kop					 updated 27-12-2016 ***/
/*** ============================================================================== ***/
/*** NAMESPACE: Crealysm::RENDERER (DX11)											***/
/*** cpp file with the DX11ShaderMgr class implementation		     				***/

#include "RENDERER/DX11/DX11shadermgr.h"

#include <CIniParser.h>
#include <ceasylogging++.h>
#include <assert.h>

#include <COMMON/Crealysm_text.h>

using namespace Crealysm::COMMON;


namespace Crealysm { namespace RENDERER
{

/**************************************************************************************/
/***								CONSTRUCTOR										***/
/*** ==> usage: when creating a DX11ShaderMgr object	 							***/
/*** ==> sets all variables in the class object to initial/ passed values			***/
/**************************************************************************************/

DX11ShaderMgr::DX11ShaderMgr(ID3D11Device *pDevice) : mDevice(pDevice), mCurrentPackId(0), mCurrentInputLayoutId(0)
{
	assert(mDevice);
	CLOG(DEBUG, "DX11RENDERER") << debug_shadermgr_created.c_str();
}

/**************************************************************************************/
/***								DESTRUCTOR										***/
/*** ==> usage: when DX11ShaderMgr object is not needed anymore					***/
/*** ==> releases stuff, COM objects, memory etc. 									***/
/**************************************************************************************/

DX11ShaderMgr::~DX11ShaderMgr()
{
	mCurrentPackId			= 0;
	mCurrentInputLayoutId	= 0;
}

/**************************************************************************************/
/***								ADD SHADERPACK									***/
/*** ==> usage: to add a shaderpack in the shadermanager (RAII proof)				***/
/*** ==> creates the pack using IO, if it fails; fallback pack is returned			***/
/**************************************************************************************/

DX11ShaderPack* DX11ShaderMgr::AddShaderPack(const std::string &pFilename, const uint pPackId, const std::string &pTarget)
{
	mShaderPacks.emplace_back();

	DX11ShaderPack *result = &mShaderPacks.back();

	if(result->CreateShaders(mDevice, mShaderFolder, pFilename, pPackId, pTarget))
	{
		return result;
	}
	mShaderPacks.pop_back();

	CLOG(DEBUG, "DX11RENDERER") << error_create_shaderpack.c_str() << " ID: " << pPackId;
	return &mFallbackPack;
}

/**************************************************************************************/
/***								LOAD SHADERS									***/
/*** ==> usage: to load the shader configuration and all shaders 					***/
/*** ==> create and/or compile shaders, and create unique inputlayouts				***/
/**************************************************************************************/

bool DX11ShaderMgr::LoadShaders(const std::string &pRootFolder, const std::string &pShaderConfigFile, const std::string &pTarget)
{
	mShaderFolder = pRootFolder + "shaders/";

	// Create fallback ShaderPack
	if(!mFallbackPack.CreateFallback(mDevice, pTarget, pRootFolder))
	{																	
		CLOG(FATAL, "DX11RENDERER") << fatal_fallback_shader.c_str();		
		return false;
	}

	// Parse shaderconfig INI file: create ShaderPacks
	Crealysm::COMMON::CIniParser packParser;
	if(!packParser.StartReading(pRootFolder + pShaderConfigFile))
	{
		CLOG(ERROR, "DX11RENDERER") << error_parse_shadercfg.c_str();		// continue with error, use fallback shaders
	}
	else
	{
		uint nrShaderPacks = 0;
		if(!packParser.CountBlocks("ShaderPack", nrShaderPacks))
		{
			CLOG(ERROR, "DX11RENDERER") << error_parse_shadercfg.c_str();	// continue with error, use fallback shaders	
		}
		else
		{
			mShaderPacks.reserve(nrShaderPacks);

			// Create all packs from ShaderConfig
			for(size_t p=0;p<nrShaderPacks;++p)
			{
				AddShaderPack(pRootFolder + pShaderConfigFile, p, pTarget);
			}
		}
	}

	// Create inputlayouts
	if(!CreateInputLayouts()) return false;

	// Set current ShaderPack and InputLayout ID's to Fallback
	mCurrentPackId = 0;
	mCurrentInputLayoutId = 0;

	CLOG(DEBUG, "DX11RENDERER") << debug_currpack_changed.c_str() << mCurrentPackId << ", " << mCurrentInputLayoutId;
	return true;
}

/**************************************************************************************/
/***								SET PACK BY ID									***/
/*** ==> usage: to set/ change the currently set ShaderPack, by numeric unique ID	***/
/*** ==> updates the current set pack and it's InputLayout, if possible 			***/
/**************************************************************************************/

bool DX11ShaderMgr::SetPackById(const uint pId)
{
	for(size_t p=0;p<mShaderPacks.size();++p)
	{
		if(mShaderPacks[p].GetIdentifier() == p)
		{
			mCurrentPackId = p;
			mCurrentInputLayoutId = mShaderPacks[mCurrentPackId].GetInputLayoutID();

			CLOG(DEBUG, "DX11RENDERER") << debug_currpack_changed.c_str() << mCurrentPackId << ", " << mCurrentInputLayoutId;
			return true;
		}
	}
	CLOG(DEBUG, "DX11RENDERER") << debug_pack_notfound.c_str() << pId;
	return false;
}

/**************************************************************************************/
/***								SET PACK BY NAME								***/
/*** ==> usage: to set/ change the currently set ShaderPack, by text identifier		***/
/*** ==> updates the current set pack, if possible 									***/
/**************************************************************************************/

bool DX11ShaderMgr::SetPackByName(const std::string &pName)
{
	for(size_t p=0;p<mShaderPacks.size();++p)
	{
		if(strcmp(mShaderPacks[p].GetIdentifierText().c_str(), pName.c_str()) == 0)
		{
			mCurrentPackId = p;
			mCurrentInputLayoutId = mShaderPacks[mCurrentPackId].GetInputLayoutID();

			CLOG(DEBUG, "DX11RENDERER") << debug_currpack_changed.c_str() << mCurrentPackId << ", " << mCurrentInputLayoutId;
			return true;
		}
	}
	CLOG(DEBUG, "DX11RENDERER") << debug_pack_notfound.c_str() << pName;
	return false;
}

/**************************************************************************************/
/***								CREATE INPUTLAYOUTS								***/
/*** ==> usage: to create all unique inputlayouts based on the shaderpacks			***/
/*** ==> creates D3D11 inputlayouts and assigns ID's to the shaderpacks 			***/
/**************************************************************************************/

bool DX11ShaderMgr::CreateInputLayouts()
{
	/*** INPUT LAYOUTS: create and store only unique ones ***/
	mInputLayouts.reserve(mShaderPacks.size()+1);					// + 1 = fallback pack
	mInputLayoutChecksums.reserve(mShaderPacks.size()+1);

	// 1st one: based on the Fallback ShaderPack
	mInputLayoutChecksums.push_back(mFallbackPack.GetInputLayoutMD5Checksum());
	mInputLayouts.push_back(mFallbackPack.GetInputLayout(mDevice));

	if(mInputLayouts.back() == nullptr)
	{
		CLOG(FATAL, "DX11RENDERER") << fatal_fallback_inplayout.c_str();
		return false;
	}
	mFallbackPack.SetInputLayoutId(0);

	// loop through the std::vector with ShaderPacks to find unique input layouts
	for(size_t v=0;v<mShaderPacks.size();++v)
	{
		// retrieve checksum of the next VS's inputlayout
		std::string checksum;
		checksum = mShaderPacks[v].GetInputLayoutMD5Checksum();

		// loop through the existing inputlayouts
		bool exists = false;
		for(size_t ex=0;ex<mInputLayouts.size();++ex)
		{
			if(checksum == mInputLayoutChecksums[ex])
			{
				mShaderPacks[v].SetInputLayoutId(ex);
				exists = true;
			}
			if(exists) break;
		}

		// new unique InputLayout found
		if(!exists)
		{
			mInputLayouts.push_back(mShaderPacks[v].GetInputLayout(mDevice));
			if(mInputLayouts.back() == nullptr)
			{
				CLOG(ERROR, "DX11RENDERER") << error_create_inplayout.c_str();
			}
			else
			{
				mInputLayoutChecksums.push_back(checksum);
				mShaderPacks[v].SetInputLayoutId(mInputLayouts.size()-1);
			}
		}
		CLOG(DEBUG, "DX11RENDERER") << debug_pack_inplayout_id.c_str() << v << " -> " << mShaderPacks[v].GetInputLayoutID();
	}
	return true;
}




/**************************************************************************************/
/***								GET NR SHADERPACKS						  CONST	***/
/*** ==> usage: to retrieve the total number of ShaderPacks							***/
/*** ==> returns the size of the corresponding std::vector							***/
/**************************************************************************************/

uint DX11ShaderMgr::GetNrShaderPacks() const
{
	return (uint)mShaderPacks.size();
}

/**************************************************************************************/
/***								GET NR INPUTLAYOUTS						  CONST	***/
/*** ==> usage: to retrieve the total number of unique inputlayouts					***/
/*** ==> returns the size of the corresponding std::vector							***/
/**************************************************************************************/

uint DX11ShaderMgr::GetNrInputLayouts() const
{
	return (uint)mInputLayouts.size();
}

/**************************************************************************************/
/***								GET CURRENT PACK ID						  CONST	***/
/*** ==> usage: to retrieve the ID of the current ShaderPack						***/
/*** ==> returns the value of the member variable									***/
/**************************************************************************************/

uint DX11ShaderMgr::GetCurrentPackId() const
{
	return mCurrentPackId;
}

/**************************************************************************************/
/***							GET CURRENT INPUTLAYOUT ID					  CONST	***/
/*** ==> usage: to retrieve the ID of the current inputlayout						***/
/*** ==> returns the value of the member variable									***/
/**************************************************************************************/

uint DX11ShaderMgr::GetCurrentInputLayoutId() const
{
	return mCurrentInputLayoutId;
}


/**************************************************************************************/
/***							GET CURRENT PACK PTR						  CONST	***/
/*** ==> usage: to retrieve a const pointer to the current ShaderPack 				***/
/*** ==> returns the pointer to the shaderpack, nullptr if pack is invalid			***/
/**************************************************************************************/

const DX11ShaderPack* DX11ShaderMgr::GetCurrentPackPtr() const
{
		return &mShaderPacks[mCurrentPackId];
}

/**************************************************************************************/
/***							GET INPUTLAYOUT PTR							  CONST	***/
/*** ==> usage: to retrieve a const pointer to the current Inputlayout 				***/
/*** ==> returns the pointer to the 'active' inputlayout							***/
/**************************************************************************************/

ID3D11InputLayout* DX11ShaderMgr::GetInputLayoutPtr() const
{
	return mInputLayouts[mCurrentInputLayoutId];
}



}
}